"use strict";

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) arr2[i] = arr[i]; return arr2; } else { return Array.from(arr); } }

var __awaiter = undefined && undefined.__awaiter || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) {
            try {
                step(generator.next(value));
            } catch (e) {
                reject(e);
            }
        }
        function rejected(value) {
            try {
                step(generator.throw(value));
            } catch (e) {
                reject(e);
            }
        }
        function step(result) {
            result.done ? resolve(result.value) : new P(function (resolve) {
                resolve(result.value);
            }).then(fulfilled, rejected);
        }
        step((generator = generator.apply(thisArg, _arguments)).next());
    });
};
function noop() {}
exports.noop = noop;
function yes() {
    return true;
}
exports.yes = yes;
function baseCompare(a, b) {
    return Number(a > b) - Number(a < b);
}
exports.baseCompare = baseCompare;
function binaryIndexOf(arr, el) {
    let compare = arguments.length <= 2 || arguments[2] === undefined ? baseCompare : arguments[2];

    let low = 0,
        high = arr.length - 1;
    while (low <= high) {
        const pivot = high + low >> 1,
              diff = compare(el, arr[pivot]);
        if (diff > 0) low = pivot + 1;else if (diff < 0) high = pivot - 1;else return pivot;
    }
    return -1;
}
exports.binaryIndexOf = binaryIndexOf;
function permissionCompare(a, b) {
    return baseCompare(a.subjectId, b.subjectId);
}
exports.permissionCompare = permissionCompare;
function permissionIndexOf(arr, subjectId) {
    return binaryIndexOf(arr, { subjectId: subjectId }, permissionCompare);
}
exports.permissionIndexOf = permissionIndexOf;
function getClassOf(node) {
    return Object.getPrototypeOf(node).constructor;
}
exports.getClassOf = getClassOf;
/**
  Run topological sort on nodes
  https://en.wikipedia.org/wiki/Topological_sorting#Kahn.27s_algorithm
 */
function topologicalSort(nodes) {
    let nameKey = arguments.length <= 1 || arguments[1] === undefined ? 'name' : arguments[1];
    let parentKey = arguments.length <= 2 || arguments[2] === undefined ? 'parent' : arguments[2];

    const nodeList = [],
          noParentList = [],
          parentMapping = new Map(),
          remainingNodes = new Set(nodes.map(n => n[nameKey]));
    const parentCounts = {};
    for (const schemaNode of nodes) {
        const name = schemaNode[nameKey],
              parentProp = schemaNode[parentKey];
        if (!name) {
            throw new Error(`No ${ nameKey } field on node = ${ schemaNode }`);
        }
        if (!parentProp) {
            noParentList.push(schemaNode);
            remainingNodes.delete(schemaNode[nameKey]);
            parentCounts[name] = 0;
        } else {
            const parents = Array.isArray(parentProp) ? parentProp : [parentProp];
            parentCounts[name] = parents.length;
            for (let parent of parents) {
                if (!parentMapping.has(parent)) {
                    parentMapping.set(parent, [schemaNode]);
                } else {
                    parentMapping.get(parent).push(schemaNode);
                }
            }
        }
    }
    while (noParentList.length) {
        const rootNode = noParentList.pop();
        nodeList.push(rootNode);
        if (parentMapping.has(rootNode[nameKey])) {
            const children = parentMapping.get(rootNode[nameKey]);
            while (children.length) {
                const child = children.pop(),
                      childName = child[nameKey];
                // parent child should be added after its last parent
                if (--parentCounts[childName] === 0) {
                    remainingNodes.delete(childName);
                    noParentList.push(child);
                }
            }
        }
    }
    if (remainingNodes.size) {
        throw new Error('Schema has a circular dependency or a missing parent! Examine definitions for ' + [].concat(_toConsumableArray(remainingNodes)).map(x => `"${ x }"`).join(', '));
    }
    return nodeList;
}
exports.topologicalSort = topologicalSort;
function flattenPromises(promises) {
    return __awaiter(this, void 0, Promise, function* () {
        const unFlattened = yield Promise.all(promises);
        const flattened = [];
        for (let i = 0, l = unFlattened.length; i < l; i++) {
            flattened.push.apply(flattened, _toConsumableArray(unFlattened[i]));
        }
        return flattened;
    });
}
exports.flattenPromises = flattenPromises;